/*
* AMD_COPYRIGHT
*
* Copyright <2003> ADVANCED MICRO DEVICES, INC. All Rights Reserved.
*
* Unless otherwise designated in writing, this software and any related 
* documentation are the confidential proprietary information of AMD. 
* THESE MATERIALS ARE PROVIDED "AS IS" WITHOUT ANY
* UNLESS OTHERWISE NOTED IN WRITING, EXPRESS OR IMPLIED WARRANTY OF ANY 
* KIND, INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, 
* NONINFRINGEMENT, TITLE, FITNESS FOR ANY PARTICULAR PURPOSE AND IN NO 
* EVENT SHALL AMD OR ITS LICENSORS BE LIABLE FOR ANY DAMAGES WHATSOEVER. 
*
* AMD does not assume any responsibility for any errors which may appear 
* in the Materials nor any responsibility to support or update the
* Materials.  AMD retains the right to modify the Materials at any time, 
* without notice, and is not obligated to provide such modified 
* Materials to you. AMD is not obligated to furnish, support, or make
* any further information available to you.
*
* END_AMD_COPYRIGHT
*/

#ifndef _DEBUG_H_
#define _DEBUG_H_

//
//     Debug level flags
//
const int DBG_NONE     = 0x00000000; // No Debugging info
const int DBG_FUNC     = 0x00000001; // Function entries 
const int DBG_WARNING  = 0x00000002; // Warning level
const int DBG_ERROR    = 0x00000004; // Error messages
const int DBG_PINS     = 0x00000008; // Enables dump of created pins in topology
const int DBG_NODES    = 0x00000010; // Enables dump of created nodes in topology
const int DBG_CONNS    = 0x00000020; // Enables dump of the connections in topology
const int DBG_STREAM   = 0x00000040; // Enables stream output.
const int DBG_POWER    = 0x00000080; // Enables power management output.
const int DBG_DMA      = 0x00000100; // Enables DMA engine output. 
const int DBG_REGS     = 0x00000200; // Enables register output.
const int DBG_PROBE    = 0x00000400; // Enables hardware probing output.
const int DBG_SYSINFO  = 0x00000800; // Enables system info output.
const int DBG_VSR      = 0x00001000; // Enables variable sample rate output.
const int DBG_PROPERTY = 0x00002000; // Enables property handler output
const int DBG_DURAUDIO = 0x00004000; // Enables DURAUDIO messages output
const int DBG_ISR      = 0x00008000; // Enables Interrupt service routine output.
const int DBG_PRD      = 0x00010000; // Enables PRD output.
const int DBG_AC97     = 0x00020000; // Enables AC97 output.
const int DBG_ALL      = 0xFFFFFFFF; // Enables all debug levels

// const int DBGFLAGS	   = 0x00000007; //  PRD, ISR, DMA and Function entries

//
//     Current Debug Level
//
 const int DBGFLAGS	   = 0x00022217; //  DBG_PROPERTY, DBG_NODES, DBG_REGS,DBG_AC97
    
//
// Define the print statement.
//
#if defined(__cplusplus)
extern "C" {
#endif // #if defined(__cplusplus)

//
// DBG is 1 in checked builds
//
#if (DBG)
/* STR_MODULENAME could cause overhead */
#define DOUT(lvl, strings)								\
		if (lvl & DBGFLAGS)								\
		{												\
			DbgPrint(STR_MODULENAME);					\
			DbgPrint##strings;							\
			DbgPrint("\n");								\
		}

#define BREAK()                     \
    DbgBreakPoint()
/*
#define DISP_NTSTAT(NTStatusMessage)					\
{														\
   LPVOID lpMessageBuffer;								\
   HMODULE Hand = LoadLibrary("NTDLL.DLL");				\
														\
   FormatMessage(										\
       FORMAT_MESSAGE_ALLOCATE_BUFFER |					\
       FORMAT_MESSAGE_FROM_SYSTEM |						\
       FORMAT_MESSAGE_FROM_HMODULE,						\
       Hand,											\
       Err,												\
       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),		\
       (LPTSTR) &lpMessageBuffer,						\
       0,												\
       NULL );											\
														\
   DOUT(DBG_ERROR, ("NTSTATUS: %s", lpMessageBuffer));	\
														\
   LocalFree( lpMessageBuffer );						\
   FreeLibrary(Hand);									\
}
*/

#else // if (!DBG)
#define DOUT(lvl, strings)
#define BREAK()
#endif // !DBG    


#if defined(__cplusplus)
}
#endif // #if defined(__cplusplus)



#endif
