/* 
 * <LIC_AMD_STD>
 * Copyright <2000-2010> ADVANCED MICRO DEVICES, INC. All Rights Reserved.
 *    
 *    
 * Unless otherwise designated in writing, this software and any related 
 * documentation are the confidential proprietary information of AMD. 
 *   
 * THESE MATERIALS ARE PROVIDED "AS IS" WITHOUT ANY UNLESS OTHERWISE 
 * NOTED IN WRITING, EXPRESS OR IMPLIED WARRANTY OF ANY KIND, INCLUDING 
 * BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT, 
 * TITLE, FITNESS FOR ANY PARTICULAR PURPOSE AND IN NO EVENT SHALL AMD 
 * OR ITS LICENSORS BE LIABLE FOR ANY DAMAGES WHATSOEVER. 
 *   
 * AMD does not assume any responsibility for any errors which may appear 
 * in the Materials nor any responsibility to support or update the Materials.  
 * AMD retains the right to modify the Materials at any time, without notice, 
 * and is not obligated to provide such modified Materials to you. AMD is 
 * not obligated to furnish, support, or make any further information 
 * available to you.
 * 
 * </LIC_AMD_STD>  
 * <CTL_AMD_STD>
 * </CTL_AMD_STD>  
 * <DOC_AMD_STD>
 * </DOC_AMD_STD>  
 */

#include "os_inc.h"

//
//	OS-Specific calls
//

void OS_DbgMsg( unsigned char *szFormat, ... )
{
#if DURAUDIO_DEBUG
	char		szBuf[ 128 ];
	va_list	arglist;
	va_start( arglist, szFormat );
	vsprintf( szBuf, szFormat, arglist );  
	va_end( arglist );
	DbgPrint( szBuf );
#endif
}

void * OS_AllocateMemory (unsigned long Size)
{
	return ExAllocatePool( NonPagedPool, Size );
}

void * OS_AllocateDMAMemory (void* pAdapterObject, unsigned long Size, unsigned long *PhysicalAddress)
{
	PHYSICAL_ADDRESS	TempPhysicalAdress;

	void *PointerToAllocatedMemory;

    if ((PointerToAllocatedMemory = HalAllocateCommonBuffer((PADAPTER_OBJECT) pAdapterObject, 
															Size, 
															&TempPhysicalAdress, 
															FALSE)) == NULL)
    {
        return FALSE;
    }

	*PhysicalAddress=TempPhysicalAdress.LowPart;

	return PointerToAllocatedMemory;
}

void OS_FreeDMAMemory (void* pAdapterObject, PVOID VirtualAddress, unsigned long PhysicalAddress, unsigned long Size)
{
	PHYSICAL_ADDRESS	TempPhysicalAdress;

	TempPhysicalAdress.LowPart=PhysicalAddress;

	HalFreeCommonBuffer( (PADAPTER_OBJECT) pAdapterObject,
						 Size,
						 TempPhysicalAdress,
						 VirtualAddress,
						 FALSE);
}

unsigned long OS_ReadPortULong(unsigned short Port)
{
	return (READ_PORT_ULONG((unsigned long *) Port));
}

void OS_WritePortULong (unsigned short Port, unsigned long Data)
{
	WRITE_PORT_ULONG ((unsigned long *) Port, Data);
}

unsigned short OS_ReadPortUShort (unsigned short Port)
{
	return (READ_PORT_USHORT((unsigned short *) Port));
}

void OS_WritePortUShort (unsigned short Port, unsigned short Data)
{
	WRITE_PORT_USHORT ((unsigned short *) Port, Data);
}

unsigned char OS_ReadPortUChar (unsigned short Port)
{
	return (READ_PORT_UCHAR((unsigned char *) Port));
}

void OS_WritePortUChar (unsigned short Port, unsigned char Data)
{
	WRITE_PORT_UCHAR ((unsigned char *) Port, Data);
}

unsigned long OS_VirtualAddress (unsigned long PhysicalAddress)
{
	PHYSICAL_ADDRESS	TempPhysicalAddress;
	unsigned long		Virtual_Addr;

	TempPhysicalAddress.LowPart=PhysicalAddress;

	if (!(Virtual_Addr = (unsigned long)MmMapIoSpace(TempPhysicalAddress, 128, MmNonCached)));
	{
        OS_DbgMsg ("OS_VirtualAddress: Map Audio IRQ reg Failed\n"); 
		return FALSE;
	}

	return Virtual_Addr ;
	
}

void OS_Free_VirtualAddress (unsigned long VirtualAddress)
{
	if (VirtualAddress)
		MmUnmapIoSpace((unsigned char *)VirtualAddress, 128);
	else
        OS_DbgMsg ("OS_Free_VirtualAddress: Virtual memory already released\n"); 
}

unsigned long OS_Get_F3BAR0_Virt(unsigned long BAR_Phys)
{
	PHYSICAL_ADDRESS	F3BAR_Phys;	
	unsigned long		BAR_Virt;

	F3BAR_Phys.LowPart=BAR_Phys;	
	F3BAR_Phys.HighPart=0;			

    // Mem-map the audio regs.
    if (!(BAR_Virt=(unsigned long) MmMapIoSpace(F3BAR_Phys, 256, MmNonCached)))
	{
        OS_DbgMsg ("OS_Get_F3BAR0_Virt: Map Audio reg Failed\n"); 
        return FALSE;
    }

	return BAR_Virt;
}

void OS_Free_F3BAR0_Virt(unsigned long BAR_Virt)
{
    // UnMem-map the audio regs.
	if (BAR_Virt)
		MmUnmapIoSpace((unsigned char *)BAR_Virt, 256);
	else
		OS_DbgMsg ("OS_Free_F3BAR0_Virt: Virtual memory already released\n"); 

}

void OS_Free (PVOID pucBuff)
{
	ExFreePool( pucBuff );
}

void OS_Sleep (unsigned long milliseconds)
{
	LARGE_INTEGER	SysTimeout; 
	NTSTATUS		RetVal;


	if( KeGetCurrentIrql() >= DISPATCH_LEVEL )
	{
		ULONG	Cnt = 20 * milliseconds;

		while( Cnt-- )
		{
			//	KeStallExecutionProcessor: counted in us
			KeStallExecutionProcessor( 50 );
		}
	}
	else
	{
        ULONGLONG       QPCStart;
        ULONGLONG       QPCEnd;
		ULONGLONG       RealWaitInTicks, RequestedWaitInNs, RealWaitInNs, RealWaitInMs;
		LARGE_INTEGER	TicksPerSecond;

		//
		//	KeDelayExecutionThread is counted in 100 ns
		//
		SysTimeout.QuadPart = (LONGLONG) (-10 * 1000 * milliseconds);
		
		RequestedWaitInNs = SysTimeout.QuadPart * 100;

		QPCEnd = QPCStart = KeQueryPerformanceCounter( &TicksPerSecond ).QuadPart;

		do
		{
			RetVal   = KeDelayExecutionThread( KernelMode, FALSE, &SysTimeout );
			QPCEnd   = KeQueryPerformanceCounter( &TicksPerSecond ).QuadPart;

			RealWaitInTicks	 = (QPCEnd - QPCStart);
			RealWaitInNs	 = (1000000000 / TicksPerSecond.QuadPart) * RealWaitInTicks;
			RealWaitInMs	 = RealWaitInNs / 1000000;
		}
		while (RealWaitInMs < milliseconds);
	}
}

//
//   Copies n bytes of audio data from the user space to the DMA buffer space
//
unsigned char OS_CopyFromUser (unsigned char *DMABuffer,  unsigned char *UserBuffer, unsigned long Count)
{
	return TRUE;
}

unsigned char OS_CopyToUser (unsigned char *UserBuffer, unsigned char *DMABuffer, unsigned long Count)
{
	return TRUE;
}

void OS_InitSpinLock (void)
{
}

void OS_SpinLock ()
{
}

void OS_SpinUnlock ()
{
}

